import pytest
from dbotu import *

import numpy as np, pandas as pd

# For Python 2 compatibility
try:
    from StringIO import StringIO
except:
    from io import StringIO

import scipy.stats, scipy.optimize
from Bio import SeqIO

@pytest.fixture
def caller():
    fasta_fh = StringIO("\n".join(['>seq1', 'AAAAAAA', '>seq2', 'AAATAAA', '>seq3', 'AATTTAA']))
    fasta_fh.seek(0)
    log_fh = StringIO()
    records = SeqIO.to_dict(SeqIO.parse(fasta_fh, 'fasta'))
    table = pd.DataFrame(np.array([[0, 10, 20], [0, 1, 2], [10, 0, 0]]), index=['seq1', 'seq2', 'seq3'], columns=['sample1', 'sample2', 'sample3'])
    return DBCaller(table, records, max_dist=15, min_fold=0.0, threshold_pval=0.001, log=log_fh)

def test_seq_abunds(caller):
    assert all(caller.seq_abunds == pd.Series([30, 10, 3], index=['seq1', 'seq3', 'seq2']))

def test_empty_otus(caller):
    assert caller.otus == []

def test_process_one(caller):
    caller._process_record(caller.seq_abunds.index[0])
    assert caller.otus[0] == OTU('seq1', 'AAAAAAA', [0, 10, 20])

def test_process_two(caller):
    for i in range(2):
        caller._process_record(caller.seq_abunds.index[i])

    assert caller.otus[0] == OTU('seq1', 'AAAAAAA', [0, 10, 20])
    assert caller.otus[1] == OTU('seq3', 'AATTTAA', [10, 0, 0])

def test_process_three(caller):
    for i in range(3):
        caller._process_record(caller.seq_abunds.index[i])

    # seq2 got merged into seq1
    assert caller.otus[0] == OTU('seq1', 'AAAAAAA', [0, 11, 22])
    assert caller.otus[1] == OTU('seq3', 'AATTTAA', [10, 0, 0])

def test_id_check():
    '''fail if there are IDs in the table that are not in the fasta'''
    fasta_fh = StringIO("\n".join(['>seq1', 'A', '>seq2', 'T']))
    fasta_fh.seek(0)
    records = SeqIO.to_dict(SeqIO.parse(fasta_fh, 'fasta'))
    table = pd.DataFrame(np.array([[0, 1], [2, 3], [4, 5]]), index=['seq1', 'seq2', 'seq3'], columns=['sample1', 'sample2'])
    with pytest.raises(RuntimeError):
        DBCaller(table, records, max_dist=15, min_fold=0.0, threshold_pval=0.001)

def test_negative_dist_cutoff():
    '''fail if trying to call with a negative distance criterion'''
    with pytest.raises(AssertionError):
        call_otus(None, None, None, -1.0, 1.0, 0.5)

def test_negative_abund_cutoff():
    '''fail if trying to call with negative abundance criterion'''
    with pytest.raises(AssertionError):
        call_otus(None, None, None, 1.0, -1.0, 0.5)

def test_negative_pval_cutoff():
    '''fail if trying to call with negative pvalue criterion'''
    with pytest.raises(AssertionError):
        call_otus(None, None, None, 1.0, 1.0, -0.5)

def test_too_large_pval_cutoff():
    '''fail if trying to call with pvalue criterion greater than 1'''
    with pytest.raises(AssertionError):
        call_otus(None, None, None, 1.0, 1.0, 1.5)



from __future__ import division
import pytest
from dbotu import *

import numpy as np, pandas as pd
import scipy.stats, scipy.optimize


def test_init():
    otu = OTU('id', 'ACGTACGT', [1, 2, 3, 4])
    assert otu.abundance == 1 + 2 + 3 + 4

def test_equal():
    otu1 = OTU('id', 'ACGT', [1, 2, 3])
    otu2 = OTU('id', 'ACGT', [1, 2, 3])
    assert otu1 == otu2


class TestAbsorb:
    def test1(self):
        otu1 = OTU('1', 'ACGT', [1, 2, 3, 4])
        otu2 = OTU('2', 'ACGT', [1, 2, 3, 4])
        otu1.absorb(otu2)
        assert all(otu1.counts == np.array([2, 4, 6, 8]))
        assert all(otu2.counts == np.array([1, 2, 3, 4]))


class TestDistanceTo:
    def test1(self):
        otu1 = OTU('1', 'AAGT', [])
        otu2 = OTU('2', 'ACGT', [])
        assert otu1.distance_to(otu2) == 1.0 / 4


class TestD:
    def test1(self):
        '''check that the difference of log likelihoods is what I computed by hand'''
        # pick some example count data
        # "a" is the count data for one OTU; "b" is the other
        a = np.array([1, 23, 2, 3])
        A = np.sum(a)
        b = np.array([10, 1797, 20, 5])
        B = np.sum(b)
        n = len(a)

        # (negative of) the log likelihood for the two models
        # 1: there are separate Poisson means for each OTU and each sample (there are 2n parameters in par)
        # 2: the means for a and b are related by a scaling factor (par[0]) (there are n + 1 parameters)
        neg_ll1 = lambda par: -sum(scipy.stats.poisson.logpmf(a, par[0: n] ** 2)) - sum(scipy.stats.poisson.logpmf(b, par[n:] ** 2))
        neg_ll2 = lambda par: -sum(scipy.stats.poisson.logpmf(a, par[1:] ** 2)) - sum(scipy.stats.poisson.logpmf(b, (par[0] * par[1:]) ** 2))

        # guess that the parameters for model 1 is just the counts for each
        par1 = np.sqrt(np.concatenate((a, b)))
        res1 = scipy.optimize.minimize(neg_ll1, x0=par1, method='Nelder-Mead')

        # guess that the parameters for model 2 are: scaled B / A and then the sum of counts for each sample scaled by A/A+B
        par2 = np.sqrt(np.insert(A / (A + B) * (a + b), 0, B / A))
        res2 = scipy.optimize.minimize(neg_ll2, x0=par2, method='Nelder-Mead')

        # using the results of those two maximizations, compute the D value numerically
        diff = -(res2.fun - res1.fun)
        numerical_D = -2.0 * diff

        # compare that numerical value against the analytical expression in DBCaller
        precision = 6
        assert round(numerical_D, precision) == round(OTU._D(a, b), precision)


class TestDistributionTestPval:
    def test1(self):
        a = np.array([1, 23, 2, 3])
        b = np.array([10, 1797, 20, 5])
        pval = OTU._distribution_test_pval(a, b)
        assert round(pval, 7) == 8.5e-5



from setuptools import find_packages, setup
import pypandoc

with open('LICENSE') as f:
    license=f.read()

setup(name='dbotu',
    version='1.1',
    description='Distribution-based OTU calling',
    long_description=pypandoc.convert('README.md', 'rst'),
    author='Scott Olesen',
    author_email='swo@alum.mit.edu',
    platforms=['any'],
    license=license,
    url='http://github.com/swo/dbotu3',
    data_files=[('input', ['data/input/counts.txt', 'data/input/seq.fa']),
        ('output', ['data/output/log.txt', 'data/output/membership.txt', 'data/output/otu.txt'])],
    classifiers=['Development Status :: 3', 'Environment :: Console',
        'Programming Language :: Python :: 3', 'Topic :: Scientific/Engineering :: Bio-Informatics'],
    packages=find_packages()
    )
